#include "StdAfx.h"
#include "util.h"
#include "DynamicArray.h"
#include "TextFile.h"
#include "CfgFile.h"
#include <memory.h>
#include <string.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const char* g_TEXMF;

const BOOL IsWindowsNT( void )
{
    OSVERSIONINFO osversion;
    memset( &osversion, 0, sizeof(OSVERSIONINFO) );
    osversion.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&osversion);
    if( osversion.dwPlatformId == VER_PLATFORM_WIN32_NT )
    {
        return TRUE;
    }
    return FALSE;
}

const BOOL IsSignaled( HANDLE h )
{
	DWORD dw = WaitForSingleObject( h, 0 );
	switch( dw )
	{
		case WAIT_ABANDONED: // mutex is not released.
			return FALSE;
		case WAIT_OBJECT_0: // signaled
			return TRUE;
		case WAIT_TIMEOUT:  // nonsignaled
			return FALSE;
		default:
			break;
	}
	return TRUE;
}

const int GetDefaultPrinterDPI( void )
{
	PRINTDLG pd;
	memset( &pd, 0, sizeof(PRINTDLG) );
	pd.lStructSize = sizeof(PRINTDLG);
	pd.Flags = PD_RETURNDEFAULT | PD_RETURNDC;
	PrintDlg( &pd );
	HDC hPrinterDC = pd.hDC;
	if( hPrinterDC == NULL ) return 0;
	int xDPI = GetDeviceCaps( hPrinterDC, LOGPIXELSX );
	int yDPI = GetDeviceCaps( hPrinterDC, LOGPIXELSY );
	if( xDPI != yDPI ) return 0;
	return xDPI;
}

BOOL IsKanji1( const unsigned char c )
{
	if( (c >= 0x81 && c <= 0x9F) || (c >= 0xE0 && c <= 0xFC) )
		return TRUE;
	else
		return FALSE;
}

void YenToSlash( char* str )
{
	/* We assume that the kanji code is "Shift JIS"! */
	int l = lstrlen( str );
	int i;
	while( (i = FindFirstChar(str, '\\')) != -1 )
	{ 
			str[i] = '/';
	}
	return;
}

void SlashToYen( char* str )
{
	int l = lstrlen( str );
	int i;
	while( (i = FindFirstChar(str, '/')) != -1 )
	{ 
			str[i] = '\\';
	}
	return;
}

BOOL IsError( const char* sz )
{
	if( sz[0] == '!' ) return TRUE;
	if( CompareFirstString(sz, "(\\end occurred inside a group") == TRUE ) return TRUE;
	return FALSE;
}

BOOL IsLaTeXError( const char* sz )
{
	if( CompareFirstString( sz, "! LaTeX Error" ) == TRUE ) return TRUE;
	return FALSE;
}

BOOL IsLaTeXWarning( const char* sz )
{
	if( CompareFirstString( sz, "LaTeX Warning" ) == TRUE ) return TRUE;
	return FALSE;
}

BOOL IsLaTeXFontWarning( const char* sz )
{
	if( CompareFirstString( sz, "LaTeX Font Warning" ) == TRUE ) return TRUE;
	else return FALSE;
}

BOOL IsTeXWarning( const char* sz )
{
	if( CompareFirstString( sz, "Overfull" ) == TRUE ) return TRUE;
	else if( CompareFirstString( sz, "Underfull" ) == TRUE ) return TRUE;
	else return FALSE;
}

BOOL IsGuishellMessage( const char* sz )
{
	if( CompareFirstString( sz, "GUI-Shell Message" ) == TRUE ) return TRUE;
	else return FALSE;
}

BOOL IsDvipsMessage( const char* sz )
{
	if( CompareFirstString( sz, "dvipsk: " ) == TRUE ) return TRUE;
	else return FALSE;
}

BOOL IsDvipsError( const char* sz )
{
	if( CompareFirstString( sz, "dvipsk: ! " ) == TRUE ) return TRUE;
	else return FALSE;
}

BOOL IsKpathseaMessage( const char* sz )
{
	if( CompareFirstString( sz, "kpathsea: " ) == TRUE ) return TRUE;
	else return FALSE;
}


BOOL GuessTEXMF( void )
{
	/* This funstion Guesses TEXMF directory from the directory of which the
	   process is in.
	   According to the directory, this function also sets texmfcnf direcroty. */
	char* module_name;
	module_name = new char[MAX_PATH];
	GetModuleFileName( NULL, module_name, MAX_PATH );
	CutoffDirectoryOneLevel( module_name );
	char* module_path = module_name;
	CutoffDirectoryOneLevel( module_path );
	char* module_parent = module_path;
	char* texmf;
	if( IsExistChildDirectory(module_parent, "texmf") )
	{
		int l = lstrlen( module_parent );
		l += lstrlen( "texmf" );
		l += 2;
		texmf = new char[l];
		wsprintf( texmf, "%s\\%s", module_parent, "texmf" );
	}
	else
	{
		if( IsExistChildDirectory(module_parent, "share\\texmf") )
		{
			int l = lstrlen( module_parent );
			l += lstrlen( "share\\texmf" );
			l += 2;
			texmf = new char[l];
			wsprintf( texmf, "%s\\%s", module_parent, "share\\texmf" );
		}
		else
		{
			delete[] module_name;
			return FALSE;
		}
	}
	int texmf_len = lstrlen( texmf );
	int texmfcnf_len = texmf_len + lstrlen( "\\web2c" );
	texmfcnf_len += 1;
	char* texmfcnf = new char[texmfcnf_len];
	wsprintf( texmfcnf, "%s\\%s", texmf, "web2c" );
	YenToSlash( texmf );
	YenToSlash( texmfcnf );
	SetEnvironmentVariable( "TEXMF", texmf );
	SetEnvironmentVariable( "TEXMFMAIN", texmf );
	SetEnvironmentVariable( "TEXMFCNF", texmfcnf );
	delete[] texmfcnf;
	delete[] texmf;
	delete[] module_name;
	return TRUE;
}

void CutoffDirectoryOneLevel( char* pathname )
{
	/* This function cuts off the file name one directory level,
	i.e, if the pathname is "a:\bbb\ccc\ddd\eee" then
	it becomes to "a:\bbb\ccc\ddd" when the function returns.
	If pathname doesn't include '\' then this function has
	nothing to do. */
	ASSERT( pathname != NULL );
	int l = lstrlen( pathname );
	int i = FindLastChar( pathname, '\\' );
	if( i != -1 ) pathname[i] = NULL;
}

BOOL IsExistChildDirectory( char* parent, char* child )
{
	/* This function searches the "child" directory in the
	"parent" directory, i.e.,
	if the parent is "a:\bbb\ccc" and child is "ddd"
	and if there is a directory of which name is "a:\bbb\ccc\ddd",
	this function returns TRUE, otherwise returns FALSE. */

	int l = lstrlen( parent );
	int c = l + lstrlen( child );
	c += 2;
	char* the_directory = new char[c];
	if( l > 0 && parent[l-1] == '\\' )
		wsprintf( the_directory, "%s%s", parent, child );
	else
		wsprintf( the_directory, "%s\\%s", parent, child );
	WIN32_FIND_DATA ffd;
	HANDLE hFind = FindFirstFile( the_directory, &ffd );
	if( hFind == INVALID_HANDLE_VALUE )
	{
		delete[] the_directory;
		return FALSE;
	}
	FindClose( hFind );
	delete[] the_directory;
	if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
		return TRUE;
	else
		return FALSE;
}

BOOL IsFileExist( const char* filename )
{
	WIN32_FIND_DATA ffd;
	HANDLE hFind = FindFirstFile( filename, &ffd );
	if( hFind == INVALID_HANDLE_VALUE ) return FALSE;
	FindClose( hFind );
	if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 ) return TRUE;
	else return FALSE;
}


BOOL GuessGSDirectory( void )
{
	CDynamicCharArray* pda = new CDynamicCharArray;
	SearchHardDisk( pda ); // list up hard disk drive on the computer.
	/* We search the "gstools" directory on the root directory of the drive
	   and under the "Program Files" directory.
	   Remark that the "Program Files" directory is specified by
	   environment variable "ProgramFilesDir". */
	char* ProgramFilesDir = new char[1];
	DWORD size = GetEnvironmentVariable( "ProgramFilesDir", ProgramFilesDir, 1 );
	if( size == 0 )
	{
		delete[] ProgramFilesDir;
		char temp[MAX_PATH+1];
		GetWindowsDirectory( temp, MAX_PATH+1 );
		for( int i = 0; i < lstrlen(temp); i++ )
		{
			if( temp[i] == '\\' )
			{
				temp[i] = NULL;
				break;
			}
		}
		ProgramFilesDir = new char[18];
		wsprintf( ProgramFilesDir, "%s\\Program Files", temp );
	}
	else
	{
		delete[] ProgramFilesDir;
		ProgramFilesDir = new char[size+2];
		GetEnvironmentVariable( "ProgramFilesDir", ProgramFilesDir, size+1 );
	}
	lstrcat( ProgramFilesDir, "\\" );

	char* gstools_dir = NULL;
	if( IsExistChildDirectory(ProgramFilesDir, "gstools") )
	{
		gstools_dir = new char[lstrlen(ProgramFilesDir) + 8];
		wsprintf( gstools_dir, "%sgstools", ProgramFilesDir );
	}
	else
	{
		for( UINT i = 0; i <= pda->GetLastIndexOfArray(); i++ )
		{
			if( IsExistChildDirectory(pda->GetMemberOf(i), "gstools") )
			{
				gstools_dir = new char[11];
				wsprintf( gstools_dir, "%sgstools", pda->GetMemberOf(i) );
				break;
			}
		}
	}
	delete[] ProgramFilesDir;

	char* aladdin_dir = NULL;
	if( IsExistChildDirectory(ProgramFilesDir, "Aladdin") )
	{
		aladdin_dir = new char[lstrlen(ProgramFilesDir) + 8];
		wsprintf( aladdin_dir, "%sAladdin", ProgramFilesDir );
	}
	else
	{
		for( UINT i = 0; i <= pda->GetLastIndexOfArray(); i++ )
		{
			if( IsExistChildDirectory(pda->GetMemberOf(i), "Aladdin") )
			{
				aladdin_dir = new char[11];
				wsprintf( aladdin_dir, "%saladdin", pda->GetMemberOf(i) );
				break;
			}
		}
	}

	if( gstools_dir == NULL && aladdin_dir == NULL )
	{
		delete pda;
		return FALSE;
	}
	/* From here, we only treat gstools_dir and aladdin_dir.
	   We first search gs* directory such gs5.01 or gstools.
	   we set PATH to that directory. */
	/* not implemented yet!!!   aladdin_dir??? */
/*
	HANDLE hFind;
	WIN32_FIND_DATA ffd;
	int gstools_size = lstrlen( gstools_dir );
	char* search_file = new char[gstools_size + 7];
	wsprintf( search_file, "%s\\gs*.*", gstools_dir );
	hFind = FindFirstFile( search_file, &ffd );
	if( hFind == INVALID_HANDLE_VALUE )
	{
		delete[] gstools_dir;
		delete[] search_file;
		delete pda;
		return FALSE;
	}
	char* path = new char[1];
	DWORD path_size = GetEnvironmentVariable( "PATH", path, 1 );
	delete[] path;
	if( path_size == 0 )
	{
		delete[] gstools_dir;
		delete[] search_file;
		delete pda;
		return FALSE;
	}
	path = new char[path_size + 1];
	GetEnvironmentVariable( "PATH", path, path_size+1 );
	path_size = lstrlen( path );
	int newpath_size = path_size;
	newpath_size += lstrlen( ffd.cFileName ) + gstools_size + 3;
	while( FindNextFile(hFind, &ffd) != FALSE )
	{
		newpath_size += lstrlen(ffd.cFileName) + gstools_size + 3;
	}
	FindClose( hFind );
	char* newpath = new char[newpath_size];
	lstrcpy( newpath, path );
	delete[] path;
	hFind = FindFirstFile( search_file, &ffd );
	if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
	{
		lstrcat( newpath, ";" );
		lstrcat( newpath, gstools_dir );
		lstrcat( newpath, "\\" );
		lstrcat( newpath, ffd.cFileName );
	}
	while( FindNextFile(hFind, &ffd) != FALSE )
	{
		if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
			continue;
		lstrcat( newpath, ";" );
		lstrcat( newpath, gstools_dir );
		lstrcat( newpath, "\\" );
		lstrcat( newpath, ffd.cFileName );
	}
	FindClose( hFind );
*/
	delete[] gstools_dir;
	delete[] aladdin_dir;
//	delete[] search_file;
//	SetEnvironmentVariable( "PATH", newpath );
//	delete[] newpath;
	delete pda;
	return TRUE;
}

BOOL GuessAcrobatDirectory( void )
{
	CDynamicCharArray* pda = new CDynamicCharArray;
	SearchHardDisk( pda ); // list up hard disk drive on the computer.
	/* We search the "gstools" directory on the root directory of the drive
	   and under the "Program Files" directory.
	   Remark that the "Program Files" directory is specified by
	   environment variable "ProgramFilesDir". */
	/* 1999/06/20:
	   Adobe Acrobat 4.0 was released on 1999/06/18.
	   Default installed directory changed from "C:\Acrobat3" to
	   "C:\Program Files\Adobe\Acrobat 4.0\Acrobat". */
	char* ProgramFilesDir = new char[1];
	DWORD size = GetEnvironmentVariable( "ProgramFilesDir", ProgramFilesDir, 1 );
	if( size == 0 )
	{
		delete[] ProgramFilesDir;
		char temp[MAX_PATH+1];
		GetWindowsDirectory( temp, MAX_PATH+1 );
		for( int i = 0; i < lstrlen(temp); i++ )
		{
			if( temp[i] == '\\' )
			{
				temp[i] = NULL;
				break;
			}
		}
		ProgramFilesDir = new char[18];
		wsprintf( ProgramFilesDir, "%s\\Program Files", temp );
	}
	else
	{
		delete[] ProgramFilesDir;
		ProgramFilesDir = new char[size+2];
		GetEnvironmentVariable( "ProgramFilesDir", ProgramFilesDir, size+1 );
	}
	lstrcat( ProgramFilesDir, "\\" );

	char* acrobat_dir = NULL;
	char search_path[MAX_PATH];
	wsprintf( search_path, "%sAcrobat*", ProgramFilesDir );
	WIN32_FIND_DATA ffd;
	HANDLE hFind = FindFirstFile( search_path, &ffd );
	UINT i;
	if( hFind != INVALID_HANDLE_VALUE )
	{
		if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
		{
			while( FindNextFile(hFind, &ffd) != FALSE )
			{
				if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
					goto find_acrobat;
				else continue;
			}
			FindClose( hFind );
			delete pda;
			return FALSE;
		}
		find_acrobat:;
		int l = lstrlen( ProgramFilesDir );
		l += lstrlen( ffd.cFileName );
		l += 2;
		acrobat_dir = new char[l];
		wsprintf( acrobat_dir, "%s%s\\", ProgramFilesDir, ffd.cFileName );
		delete[] ProgramFilesDir;
		ProgramFilesDir = NULL;
		goto set_path;
	}
	wsprintf( search_path, "%sAdobe\\Acrobat*.*", ProgramFilesDir );
	hFind = FindFirstFile( search_path, &ffd );
	if( hFind != INVALID_HANDLE_VALUE )
	{
		if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
		{
			while( FindNextFile(hFind, &ffd) != FALSE )
			{
				if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
					goto find_acrobat_4;
				else continue;
			}
			FindClose( hFind );
			delete pda;
			return FALSE;
		}
		find_acrobat_4:;
		int l = lstrlen( ProgramFilesDir );
		l += lstrlen( ffd.cFileName );
		l += 6; // "Adobe\\"
		l += 2; // one "\\" and NULL
		acrobat_dir = new char[l];
		wsprintf( acrobat_dir, "%sAdobe\\%s\\", ProgramFilesDir, ffd.cFileName );
		delete[] ProgramFilesDir;
		ProgramFilesDir = NULL;
		goto set_path;
	}
	if( ProgramFilesDir != NULL ) delete[] ProgramFilesDir;
	for( i = 0; i <= pda->GetLastIndexOfArray(); i++ )
	{
		wsprintf( search_path, "%sAcrobat*", pda->GetMemberOf(i) );
		hFind = FindFirstFile( search_path, &ffd );
		if( hFind != INVALID_HANDLE_VALUE )
		{
			if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
			{
				while( FindNextFile(hFind, &ffd) != FALSE )
				{
					if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
						goto find_acrobat_root;
					else continue;
				}
				FindClose( hFind );
				delete pda;
				return FALSE;
			}
			find_acrobat_root:;
			int l = lstrlen( ffd.cFileName );
			l += 5;
			acrobat_dir = new char[l];
			wsprintf( acrobat_dir, "%s%s\\", pda->GetMemberOf(i), ffd.cFileName );
			goto set_path;
		}
		else
			continue;
	}
	delete pda;
	return FALSE;

	set_path:;
	wsprintf( search_path, "%s*.*", acrobat_dir );
	hFind = FindFirstFile( search_path, &ffd );
	if( hFind == INVALID_HANDLE_VALUE )
	{
		delete[] acrobat_dir;
		delete pda;
		return FALSE;
	}

	char* path = new char[1];
	DWORD path_size = GetEnvironmentVariable( "PATH", path, 1 );
	delete[] path;
	if( path_size == 0 )
	{
		delete[] acrobat_dir;
		delete pda;
		return FALSE;
	}
	path = new char[path_size + 1];
	GetEnvironmentVariable( "PATH", path, path_size+1 );
	path_size = lstrlen( path );
	int newpath_size = path_size;
	int acrobat_dir_size = lstrlen( acrobat_dir );
	if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
	{
		newpath_size += acrobat_dir_size;
		newpath_size += lstrlen( ffd.cFileName ) + 2;
	}
	while( FindNextFile(hFind, &ffd) != FALSE )
	{
		if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
		{
			newpath_size += acrobat_dir_size;
			newpath_size += lstrlen( ffd.cFileName ) + 2;
		}
	}
	FindClose( hFind );
	char* newpath = new char[newpath_size];
	lstrcpy( newpath, path );
	hFind = FindFirstFile( search_path, &ffd );
	if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 )
	{
		if( lstrcmp(ffd.cFileName,".") != 0 && lstrcmp(ffd.cFileName,"..") != 0 )
		{
			lstrcat( newpath, ";" );
			lstrcat( newpath, acrobat_dir );
			lstrcat( newpath, ffd.cFileName );
		}
	}
	while( FindNextFile(hFind, &ffd) != FALSE )
	{
		if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
			continue;
		if( lstrcmp(ffd.cFileName,".") != 0 && lstrcmp(ffd.cFileName,"..") != 0 )
		{
			lstrcat( newpath, ";" );
			lstrcat( newpath, acrobat_dir );
			lstrcat( newpath, ffd.cFileName );
		}
	}
	FindClose( hFind );
	delete[] acrobat_dir;
	delete[] path;
	SetEnvironmentVariable( "PATH", newpath );
	delete[] newpath;
	delete pda;
	return TRUE;
}

BOOL SetExecutablePath( void )
{
	char module_name[MAX_PATH];
	GetModuleFileName( NULL, module_name, MAX_PATH );
	CutoffDirectoryOneLevel( module_name );
	char* path;
	path = new char[1];
	DWORD path_size = GetEnvironmentVariable( "PATH", path, 1 );
	delete[] path;
	DWORD newpath_size = path_size + lstrlen(module_name) + 2;
	path = new char[newpath_size];
	GetEnvironmentVariable( "PATH", path, newpath_size );
	lstrcat( path, ";" );
	lstrcat( path, module_name );
	SetEnvironmentVariable( "PATH", path );
	delete[] path;
	return TRUE;
}

void SearchHardDisk( CDynamicCharArray* pda )
{
	char root_path[4];
	lstrcpy( root_path, "a:\\" );
	for( char c = 'a'; c <= 'z'; c++ )
	{
		root_path[0] = c;
		if( GetDriveType(root_path) == DRIVE_FIXED )
		{
			*pda << root_path;
		}
		else
			continue;
	}
	return;
}

void SearchCDROM( CDynamicCharArray* pda )
{
	char root_path[4];
	lstrcpy( root_path, "a:\\" );
	for( char c = 'a'; c <= 'z'; c++ )
	{
		root_path[0] = c;
		if( GetDriveType(root_path) == DRIVE_CDROM )
		{
			*pda << root_path;
		}
		else
			continue;
	}
	return;
}

int FindLastChar( const char* str, char c )
{
	/* This function checks whether the char c exists in the str.
	   If c exists, this function returns the LAST index of c.
	   If not exists, return -1; */
	int ret = -1;
	BOOL prev_is_kanji1 = FALSE;
	int l = lstrlen( str );
	for( int i = 0; i < l; i ++ )
	{
		if( prev_is_kanji1 == TRUE )
		{// we must skip the 2nd byte of Shift-JIS.
		 // the next character of 2nd byte of Shift-JIS is always not a 2nd byte of Kanji.
			prev_is_kanji1 = FALSE;
			continue;
		}
		else // prev_is_kanji == FALSE
		{// the previous character is not Shift-JIS first code. 
			if( str[i] == c )
			{
				ret = i;
				continue;
			}
			else if( IsKanji1(str[i]) == TRUE )
			{
				prev_is_kanji1 = TRUE;
				continue;
			}
			else
				continue;
		}
	}
	return ret;
}

int FindFirstChar( const char* str, char c )
{
	/* This function checks whether the char c exists in the str.
	   If c exists, this function returns the FIRST index of c.
	   If not exists, return -1; */
	BOOL prev_is_kanji1 = FALSE;
	int l = lstrlen( str );
	for( int i = 0; i < l; i ++ )
	{
		if( prev_is_kanji1 == TRUE )
		{// we must skip the 2nd byte of Shift-JIS.
		 // the next character of 2nd byte of Shift-JIS is always not a 2nd byte of Kanji.
			prev_is_kanji1 = FALSE;
			continue;
		}
		else // prev_is_kanji == FALSE
		{// the previous character is not Shift-JIS first code. 
			if( str[i] == c )
			{
				return i;
				continue;
			}
			else if( IsKanji1(str[i]) == TRUE )
			{
				prev_is_kanji1 = TRUE;
				continue;
			}
			else
				continue;
		}
	}
	return -1;
}

int GetErrorLine( CDynamicCharArray* src, int* line, char* file, int size_buffer )
{
	// not used.
	ASSERT( FALSE );
	ASSERT( src != NULL && line != NULL && file != NULL );
	return 0;
}

/*********************************************************************
In the texmf.cnf file, there exist the following lines:
-----------
% It's probably inadvisable to change these. At any rate, we must have:
% 45 < error_line      < 255;
% 30 < half_error_line < error_line - 15;
% 60 <= max_print_line;
% These apply to Metafont and MetaPost as well.
error_line = 79
half_error_line = 50
max_print_line = 79
----------
Here, 
error_line : width of context lines on terminal error messages
half_error_line : width of first lines of contexts in terminal error messages
max_print_line : width of longest text lines output
SEE "TeX: The Program section 11".
**********************************************************************/

/* TeX internal constants. */
static int error_line = 79;
static int half_error_line = 50;
static int max_print_line = 79;
static BOOL b_get_tex_constants = FALSE;

inline int get_max_print_line( void )
{
	if( !b_get_tex_constants ) GetTeXInternalConstants();
	return max_print_line;
}

void GetTeXInternalConstants( void )
{
	if( b_get_tex_constants ) return;
	CString strTEXMFCNF = GetTEXMFCNF();
	CCfgFile texmfcnf( strTEXMFCNF, READONLY );
	if( texmfcnf.IsOpenedFile() == FALSE ) return;
	char* temp;
	temp = texmfcnf.GetValue( "error_line" );
	if( temp != NULL ) error_line = atoi( temp );
	else return;
	temp = texmfcnf.GetValue ( "half_error_line" );
	if( temp != NULL ) half_error_line = atoi( temp );
	else return;
	temp = texmfcnf.GetValue( "max_print_line" );
	if ( temp != NULL ) max_print_line = atoi( temp );
	else return;
	b_get_tex_constants = TRUE;
	return;
}

BOOL NeedStop( const char* line )
{
	if( IsError(line) || IsLaTeXError(line) || IsTeXWarning(line)
		|| IsLaTeXFontWarning(line) || IsLaTeXWarning(line) ) return TRUE;
	return FALSE;
}

int IndexOfNonSpace( const char* line )
{
	int ret = 0;
	while( line[ret] == ' ' ) ret++;
	return ret - 1;
}

ERROR_INFO& CreateErrorInformation( CDynamicCharArray* src, CDynamicCharArray* info, BOOL FontWarning )
{
	/* REMARK: give as info the CDynamicCharArray which has NO DATA! */
	ASSERT( src != NULL );
	ASSERT( info != NULL );
	const UINT last_index = src->GetLastIndexOfArray();
	CDynamicCharArray file; // `file' is treated as a stack.
					        // the last member always current scanning file.
							// after checking error information, the `file'
							// will have no members.
	int current_file = 0;

	static ERROR_INFO err_info;
	err_info.error_exist = FALSE;
	err_info.tex_error = FALSE;
	err_info.cite_warning = FALSE;
	err_info.label_change = FALSE;
	err_info.undef_label = FALSE;

	err_info.error_exist = 0;
#ifdef _DEBUG
	FILE* fp = fopen( "test_errorinfo.log", "w" );
#endif
	for( UINT i = 0; i < last_index; i++ ) // don't check the last line.
	{
		UINT current_i = i;
		int l = lstrlen( (*src)[i] );
		/* If l >= max_print_line, we assume that the current line is continuing to
		   the next line. */
		if( l >= max_print_line && NeedStop((*src)[i]) ) l += lstrlen( (*src)[i+1] );
		char* cur_line = new char[l+1];
		lstrcpy( cur_line, (*src)[i] );
		if( l >= max_print_line && NeedStop((*src)[i]) ) lstrcat( cur_line, (*src)[i+1] );
		char* tofree = cur_line;
		if( IsGuishellMessage(cur_line) )
		{
			delete[] tofree;
			continue;
		}
#ifdef _DEBUG
		fprintf( fp, "%s\n", cur_line );
#endif

		/* !!!NOTICE!!!
		  In the following routine, we use the index i.
		  Set "i" to the last line of current scanning. */
		if( IsLaTeXError(cur_line) == TRUE )
		{
			BOOL first_error_line = TRUE;
			err_info.error_exist |= TRUE;
			err_info.tex_error |= TRUE;
			const char* err_file_c;
			if( file.IsElementExist() == TRUE )
				err_file_c = file[file.GetLastIndexOfArray()];
			else
				err_file_c = "Unknown";
			int size_err_file = lstrlen( err_file_c );
			char* err_file = new char[size_err_file+1];
			lstrcpy( err_file, err_file_c );
			SlashToYen( err_file );
			int size = 4; // ": " and "()".
			size += lstrlen( err_file );
			size += lstrlen( cur_line + 2 );
			/* make "i" indexes the last line of error message. */
			while( i < last_index )
			{
				int cur_len = lstrlen( (*src)[i] );
				if( first_error_line && cur_len >= half_error_line )
				{
					first_error_line = FALSE;
					i++;
				}
				else if( !first_error_line && cur_len >= error_line )
				{
					i++;
				}
				else break;
			}
			/* skip lines while doesn't "l." or "<*>" appear. */
			while( i < last_index )
			{
				if( CompareFirstString( (*src)[i], "l." ) == TRUE
					|| CompareFirstString( (*src)[i], "<*>" ) == TRUE ) break;
#ifdef _DEBUG
				fprintf( fp, "%s\n", (*src)[i] );
#endif
				if( i != current_i && NeedStop( (*src)[i] ) )
				{
					i--;
					break;
				}
				else
				{
					i++;
					continue;
				}
			}
			const char* line_info_c = (*src)[i];
			int size_line_info = lstrlen(line_info_c);
			char* line_info = new char[size_line_info+1];
			lstrcpy( line_info, line_info_c );
			char* info_data;
			int m = FindFirstChar( line_info + 2, ' ' );
			if( m != -1 ) line_info[m+2] = NULL;
			size += lstrlen( line_info + 2 );
			info_data = new char[size+1];
			wsprintf( info_data, "%s(%s): %s", err_file, line_info+2, cur_line+2 );
			(*info) << info_data;
			delete[] info_data;
			delete[] line_info;
			delete[] err_file;
			if( i < last_index - 1 && IndexOfNonSpace( (*src)[i+1] ) + 1 >= size_line_info ) i++;
		}
		else if( IsError(cur_line) == TRUE )
		{
			BOOL first_error_line = TRUE;
			err_info.error_exist |= TRUE;
			err_info.tex_error |= TRUE;
			const char* err_file_c;
			if( file.IsElementExist() == TRUE )
				err_file_c = file[file.GetLastIndexOfArray()];
			else
				err_file_c = "Unknown";
			int size_err_file = lstrlen( err_file_c );
			char* err_file = new char[size_err_file+1];
			lstrcpy( err_file, err_file_c );
			SlashToYen( err_file );
			int size = 4; // ": " and "()".
			size += lstrlen( err_file );
			size += lstrlen( cur_line + 2 );
			/* make "i" indexes the last line of error message. */
			while( i < last_index )
			{
				int cur_len = lstrlen( (*src)[i] );
				if( first_error_line && cur_len >= half_error_line )
				{
					first_error_line = FALSE;
					i++;
				}
				else if( !first_error_line && cur_len >= error_line )
				{
					i++;
				}
				else break;
			}
			/* skip lines while doesn't "l." or "<*>" appear. */
			while( i < last_index )
			{
				if( CompareFirstString( (*src)[i], "l." ) == TRUE
					|| CompareFirstString( (*src)[i], "<*>" ) == TRUE ) break;
#ifdef _DEBUG
				fprintf( fp, "%s\n", (*src)[i] );
#endif
				if( i != current_i && NeedStop( (*src)[i] ) )
				{
					i--;
					break;
				}
				else
				{
					i++;
					continue;
				}
			}
			const char* line_info_c = (*src)[i];
			int size_line_info = max( 2, lstrlen(line_info_c) );
			char* line_info = new char[size_line_info+1];
			lstrcpy( line_info, line_info_c );
			char* info_data;
			if( line_info[0] != 'l' )
			{
				lstrcpy( line_info, "??" );
				size += 2;
				info_data = new char[size+1];
				wsprintf( info_data, "%s(%s): %s", err_file, line_info, cur_line+2 );
			}
			else
			{
				int m = FindFirstChar( line_info + 2, ' ' );
				if( m != -1 ) line_info[m+2] = NULL;
				size += lstrlen( line_info + 2 );
				info_data = new char[size+1];
				wsprintf( info_data, "%s(%s): %s", err_file, line_info+2, cur_line+2 );
			}
			(*info) << info_data;
			delete[] info_data;
			delete[] line_info;
			delete[] err_file;
			if( i < last_index - 1 && IndexOfNonSpace( (*src)[i+1] ) + 1 >= size_line_info ) i++;
		}
		else if( IsLaTeXWarning(cur_line) == TRUE )
		{
			err_info.error_exist |= TRUE;
			if( CompareFirstString(cur_line+15, "Citation") == TRUE )
				err_info.cite_warning |= TRUE;
			if( CompareFirstString(cur_line+15, "Label(s) may have changed.") == TRUE )
				err_info.label_change |= TRUE;
			if( CompareFirstString(cur_line+15, "There were undefined reference") == TRUE )
				err_info.undef_label |= TRUE;
			const char* err_file_c;
			if( file.IsElementExist() == TRUE )
				err_file_c = file[file.GetLastIndexOfArray()];
			else
				err_file_c = "Unknown";
			int size_err_file = lstrlen( err_file_c );
			char* err_file = new char[size_err_file+1];
			lstrcpy( err_file, err_file_c );
			SlashToYen( err_file );
			int size = 4; // ": " and "()".
			size += lstrlen( err_file );
			int n = FindFirstChar( cur_line, ':' );
			cur_line = cur_line + n + 1;
			int s = FindLastChar( cur_line, ' ' );
			char* line_info = cur_line + s + 1;
			cur_line[s] = NULL;
			int p = FindLastChar( line_info, '.' );
			if( p != -1 ) line_info[p] = NULL;
			size += lstrlen( cur_line );
			size += lstrlen( line_info );
			while( i < last_index )
			{
				int cur_len = lstrlen( (*src)[i] );
				if( cur_len >= max_print_line )	i++;
				else break;
			}
#ifdef _DEBUG
			fprintf( fp, "%s\n", (*src)[i] );
#endif
			if( i != current_i && NeedStop((*src)[i]) ) i--;

			if( atoi(line_info) == 0 )
			{
/*
				while( i < last_index
					&& lstrcmp((*src)[i], "") != 0
					&& CompareFirstString( (*src)[i], "[" ) == FALSE
					&& CompareFirstString( (*src)[i], ")" ) == FALSE
					&& IsLaTeXError( (*src)[i] ) == FALSE
					&& IsError( (*src)[i] ) == FALSE
					&& IsTeXWarning( (*src)[i] ) == FALSE
					&& IsLaTeXFontWarning( (*src)[i] ) == FALSE )
				{// skip TeX's error information lines.
				 // which is the non-empty lines before empty one appers.
					i++;
				}
				i--;
*/
				delete[] err_file;
				delete[] tofree;
				continue;
			}
			char* info_data = new char[size+1];
			wsprintf( info_data, "%s(%s): %s", err_file, line_info, cur_line );
			(*info) << info_data;
			delete[] info_data;
			delete[] err_file;
		}
		else if( IsLaTeXFontWarning(cur_line) == TRUE )
		{
			if( FontWarning == FALSE )
			{ // don't mind the font warning!
				while( i < last_index )
				{
					int cur_len = lstrlen( (*src)[i] );
					if( cur_len >= max_print_line )	i++;
					else break;
				}
				if( i != current_i && NeedStop((*src)[i]) ) i--;
/*
				while( i < last_index && lstrcmp((*src)[i], "") != 0
					&& CompareFirstString( (*src)[i], "["  ) == FALSE
					&& IsLaTeXError( (*src)[i] ) == FALSE
					&& IsError( (*src)[i] ) == FALSE
					&& IsLaTeXWarning( (*src)[i] ) == FALSE
					&& IsTeXWarning( (*src)[i] ) == FALSE )
				{// skip TeX's error information lines.
				 // which is the non-empty lines before empty one appers.
					i++;
				}
				i--;
*/
				delete[] tofree;
				continue;
			}
			err_info.error_exist |= TRUE;
			const char* err_file_c;
			if( file.IsElementExist() == TRUE )
				err_file_c = file[file.GetLastIndexOfArray()];
			else
				err_file_c = "Unknown";
			int size_err_file = lstrlen( err_file_c );
			char* err_file = new char[size_err_file+1];
			lstrcpy( err_file, err_file_c );
			SlashToYen( err_file );
			int size = 4; // ": " and "()".
			size += lstrlen( err_file );
			int n = FindFirstChar( cur_line, ':' );
			cur_line = cur_line + n + 1; // Font Warning;
			const char* line_info_c = (*src)[i+1];
			int size_line_info = lstrlen(line_info_c);
			char* line_info = new char[size_line_info+1];
			lstrcpy( line_info, line_info_c );
			int p = FindLastChar( line_info, '.' );
			if( p != -1 ) line_info[p] = NULL;
			int s = FindLastChar( line_info, ' ' );
			char* line_info_l = line_info + s + 1; // the line;
			size += lstrlen( cur_line );
			size += lstrlen( line_info_l );
			if( atoi(line_info_l) == 0 )
			{
				while( i < last_index )
				{
					int cur_len = lstrlen( (*src)[i] );
					if( cur_len >= max_print_line )	i++;
					else break;
				}
				if( i != current_i && NeedStop((*src)[i]) ) i--;
/*
				while( i < last_index
					&& lstrcmp((*src)[i], "") != 0
					&& IsLaTeXError( (*src)[i] ) == FALSE
					&& IsError( (*src)[i] ) == FALSE
					&& IsLaTeXWarning( (*src)[i] ) == FALSE
					&& IsTeXWarning( (*src)[i] ) == FALSE )
				{// skip TeX's error information lines.
				 // which is the non-empty lines before empty one appers.
					i++;
				}
				i--;
*/
				delete[] line_info;
				delete[] err_file;
				delete[] tofree;
				continue;

			}
			char* info_data = new char[size+1];
			wsprintf( info_data, "%s(%s): %s", err_file, line_info_l, cur_line );
			(*info) << info_data;
			delete[] info_data;
			delete[] line_info;
			delete[] err_file;
			while( i < last_index )
			{
				int cur_len = lstrlen( (*src)[i] );
				if( cur_len >= max_print_line )	i++;
				else break;
			}
			if( i != current_i && NeedStop((*src)[i]) ) i--;
/*
			while( i < last_index
				&& lstrcmp( (*src)[i], "" ) != 0 
				&& IsLaTeXError( (*src)[i] ) == FALSE
				&& IsError( (*src)[i] ) == FALSE
				&& IsLaTeXWarning( (*src)[i] ) == FALSE
				&& IsTeXWarning( (*src)[i] ) == FALSE )
			{// skip TeX's error information lines.
			 // which is the non-empty lines before empty one appers.
				i++;
			}
			i--;
*/
		}
		else if( IsTeXWarning(cur_line) == TRUE )
		{
			err_info.error_exist |= TRUE;
			const char* err_file_c;
			if( file.IsElementExist() == TRUE )
				err_file_c = file[file.GetLastIndexOfArray()];
			else
				err_file_c = "Unknown";
			int size_err_file = lstrlen( err_file_c );
			char* err_file = new char[size_err_file+1];
			lstrcpy( err_file, err_file_c );
			SlashToYen( err_file );
			int size = 4; // ": " and "()".
			size += lstrlen( err_file );
			
			int n = FindFirstChar( cur_line, ')' );
			int s = FindLastChar( cur_line, ' ' );
			cur_line[n+1] = NULL; // warning message.
			char* line_info = cur_line + s + 1;
			int h = FindFirstChar( line_info, '-' );
			if( h != -1 ) line_info[h] = NULL;
			size += lstrlen( cur_line );
			size += lstrlen( line_info );
			char* info_data = new char[size+1];
			wsprintf( info_data, "%s(%s): %s", err_file, line_info, cur_line );
			(*info) << info_data;
			delete[] info_data;
			delete[] err_file;
			/* "i" indexes the line which contain "Overfull ...".
			   Next line describes the token list.
			   set "i" to point the first line of the token list. */
			i++;
			while( i < last_index )
			{
				int cur_len = lstrlen( (*src)[i] );
				if( cur_len >= max_print_line )	i++;
				else break;
			}
			if( i != current_i && NeedStop((*src)[i]) ) i--;

/*
			while( i < last_index && lstrcmp((*src)[i], "") != 0 &&
				CompareFirstString( (*src)[i], "["  ) == FALSE
				&& IsLaTeXError( (*src)[i] ) == FALSE
				&& IsError( (*src)[i] ) == FALSE
				&& IsLaTeXWarning( (*src)[i] ) == FALSE
				&& IsLaTeXFontWarning( (*src)[i] ) == FALSE )
			{// skip TeX's error information lines.
			 // which is the non-empty lines before empty one appers.
				i++;
			}
			i--; // 1999/06/23.
*/
		}
		else // not error/warning lines.
		{
			trying:;
			int first_close = FindFirstChar( cur_line, ')' );
			int first_open  = FindFirstChar( cur_line, '(' );
			if( first_open == -1 && first_close == -1 )
			{ // there aren't '(' or ')'.
			  // we have nothing to do.
				delete[] tofree;
				continue;
			}
			else if( -1 < first_close )
			{
				if( first_open == -1 || first_close < first_open )
				{ // close a previous file.
					if( file.GetLastIndexOfArray() != 0 )
					{
#ifdef _DEBUG
						fprintf( fp, "%s\n", file.GetMemberOf(file.GetLastIndexOfArray()) );
#endif
						file.DeleteLastMember();
					}
					cur_line += first_close + 1;
					goto trying;
				}
				else if( -1 < first_open && first_open < first_close )
				{ // open new file.
		opening:;
					char* newfilename = new char[MAX_PATH];
					wsprintf( newfilename, "%s", cur_line + first_open + 1 );
					int s = FindFirstChar( newfilename, ' ' );
					int t = FindFirstChar( newfilename, ')' );
					if( s != -1 && t != -1 ) s = min( s, t );
					else if( s == -1 || t == -1 ) s = max( s, t );
					if( s != -1 ) newfilename[s] = NULL;
					if( CompareFirstString( newfilename, "./" ) == TRUE )
						file << (newfilename + 2);
					else file << newfilename;
#ifdef _DEBUG
					fprintf( fp, "%s\n", newfilename );
#endif
					delete[] newfilename;
					cur_line += first_open + s + ((s==-1)?2:1);
					goto trying;
				}
			}
			else if( -1 < first_open )
				goto opening;
		}
		delete[] tofree;
	} // end of loop. check each line;
#ifdef _DEBUG
	fclose( fp );
#endif
	return err_info;
}

BOOL ContainString( const char* filename, const char* str )
{
	/* This function checks the "filename" file contain a line
	   which starts by "str". */
	CTextFile txt( filename );
	if( txt.IsOpenedFile() == FALSE ) return FALSE;
	UINT lines = txt.GetNumberOfLines();
	for( UINT i = 1; i <= lines; i++ )
	{
		const char* line = txt.GetTheLine( i );
		if( CompareFirstString(line, str) == TRUE )
		{
			return TRUE;
		}
	}
	return FALSE;
}

CString& GetTEXMFCNF( void )
{
	static CString texmfcnf_file = "";
	
	char* texmfcnf = new char[1];
	DWORD texmfcnf_size = GetEnvironmentVariable( "TEXMFCNF", texmfcnf, 1 );
	if( texmfcnf_size == 0 )
	{
		return texmfcnf_file;
	}
	delete[] texmfcnf;
	texmfcnf = new char[texmfcnf_size+1];
	GetEnvironmentVariable( "TEXMFCNF", texmfcnf, texmfcnf_size + 1 );
	CDynamicCharArray texmfcnf_dir;
	UINT i, j;
	j = 0;
	for( i = 0; i < (UINT)lstrlen(texmfcnf); i++ )
	{
		if( texmfcnf[i] == ';' )
		{
			char* temp = new char[i-j+1];
			memcpy( temp, texmfcnf + j, i-j );
			temp[i-j] = NULL;
			texmfcnf_dir << temp;
			delete[] temp;
			j = i + 1;
		}
	}
	texmfcnf_dir << texmfcnf + j;
	delete[] texmfcnf;

	/* Try to find texmf.cnf */
	char* try_file;
	for( i = 0; i <= texmfcnf_dir.GetLastIndexOfArray(); i++ )
	{
		int l = lstrlen( texmfcnf_dir[i] );
		l += 11;
		try_file = new char[l];
		wsprintf( try_file, "%s\\texmf.cnf", texmfcnf_dir[i] );
		SlashToYen( try_file );
		WIN32_FIND_DATA ffd;

		HANDLE hFind = FindFirstFile( try_file, &ffd );
		if( hFind == INVALID_HANDLE_VALUE )
		{
			delete[] try_file;
			FindClose( hFind );
			continue;
		}
		else 
		{
			FindClose( hFind );
			texmfcnf_file = try_file;
			delete[] try_file;
			break;
		}
	}
	return texmfcnf_file;
}

char* ReplaceString( char* original, const char* old_word, const char* new_word )
{
	/* Search old_word from original.
	   if find it, replace it by new_word.
	   Otherwize, return NULL.
	   This doesn't check the buffer is enough. */
	char* s = strstr( original, old_word );
	if( s == NULL || s == original ) return NULL;
	
	int len_org = lstrlen( original );
	int len_old = lstrlen( old_word );
	int len_new = lstrlen( new_word );
	memmove( s + len_new, s + len_old, len_org - len_old - (s-original) + 1 );
	memcpy( s, new_word, len_new );
	return original;
}

void CopyTEXMF( char* dest )
{
	char* texmf = new char[1];
	DWORD size = GetEnvironmentVariable( g_TEXMF, texmf, 1 );
	delete[] texmf;
	texmf = new char[size+1];
	GetEnvironmentVariable( g_TEXMF, texmf, size+1 );
	lstrcpy( dest, texmf );
	delete[] texmf;
}

CString& PickupLine( const char* file_name, const char* start_string )
{
	static CString pickup;
	pickup = "";
	CTextFile text;
	text.OpenFile( file_name, READONLY );
	if( text.IsOpenedFile() == FALSE ) return pickup;
	int len = lstrlen( start_string );
	BOOL all_mode = FALSE;

	for( unsigned int i = 1; i < text.GetNumberOfLines(); i++ )
	{
		char citation[256];
		if( CompareFirstString(text.GetTheLine(i), start_string) )
		{
			lstrcpy( citation, text.GetTheLine(i)+len );
			int p = FindLastChar(citation,'}');
			if( p != -1 ) citation[p] = NULL;
			if( strstr(pickup,citation) == NULL && all_mode == FALSE )
			{
				pickup += citation;
				pickup += ",";
			}
			if( lstrcmp( citation, "*" ) == 0 ) all_mode = TRUE;
		}
	}
	return pickup;
}

DWORD GetStringCheckSum( const char* str )
{
	DWORD ret = 0;
	int i = 0;
	while( str[i] != NULL )
	{
		ret += str[i];
		i++;
	}
	return ret;
}

